"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

"""This module contains all supported obligation types."""


from nagini_translation.translators.obligation.types.must_terminate import (
    MustTerminateObligationInstance,
    MustTerminateObligation,
)
